/* $Id: isdnl2.h,v 1.4 2001/03/15 09:03:35 kai Exp $
 *
 * Layer 2 defines
 *
 * This file is (c) under GNU General Public License
 *
 */

#define RR     0x01
#define RNR    0x05
#define REJ    0x09
#define SABME  0x6f
#define SABM   0x2f
#define DM     0x0f
#define UI     0x03
#define DISC   0x43
#define UA     0x63
#define FRMR   0x87
#define XID    0xaf

#define CMD    0
#define RSP    1

#define LC_FLUSH_WAIT 1
